#open "genlex";;
#open "files";;
#open "unif";;
#open "res";;

let lexer = make_lexer ["~";"&";"|";"Ex"; "Al";"(";")";"<->";"->";";"];;

type Value =
  Term of term
| Prop of prop
| Cst of cst_info
;;

let name_tbl = (myhashtbl__new 23 myhashtbl__Standard 
                                  myhashtbl__Standard 
                   : (string, Value) myhashtbl__t);;

let parse_nterm env depth n str = 
  let rec fn = function 
    [< 'Ident s ; str>] -> (
       try BVar (depth - (assoc s env)) with Not_found -> 
       try match myhashtbl__find name_tbl s with
         Term t -> t
       | Cst ({Cst_arity = p; Cst_type = Fun_sym} as n) -> Fun(n, gn p str)
       | _ -> raise (Failure ("can not use \""^s^
                                 "\" to construct a term"))
       with Not_found -> 
         raise (Failure ("identifier \""^s^"\" undefined")))
  | [< 'Kwd "("; fn t; 'Kwd ")" >] -> t 

  and gn = function
    0 -> (function [< >] -> [])
  | n -> (function [< fn t; (gn (n-1)) l >] -> t::l)

  in gn n str 
;;

let rec parse_atom env depth = function
    [< 'Ident s ; str>] -> (
       try assoc s env;raise (Failure ("can not use \""^s^
                                 "\" to construct a proposition"))
       with Not_found -> 
       try match myhashtbl__find name_tbl s with
         Prop t -> t
       | Cst ({Cst_arity = p; Cst_type = Pred_sym} as n) -> 
           Atom (Fun(n, parse_nterm env depth p str))
       | _ -> raise (Failure ("can not use \""^s^
                                 "\" to construct a proposition"))
       with Not_found -> 
         raise (Failure ("identifier \""^s^"\" undefined")))
|   [< 'Kwd ("Ex"|"Al" as q); 'Ident s; 
       (parse_atom ((s,depth+1)::env) (depth+1)) p >] ->
       (match q with
         "Al" -> Al p
       | "Ex" -> Ex p
       | _ -> raise (Failure "bug in parse_atom"))
|   [< 'Kwd "~"; (parse_atom env depth) p >] -> Non p
|   [< 'Kwd "("; (parse_ar env depth) p; 'Kwd ")" >] -> p

and parse_con env depth = function
    [< (parse_atom env depth) p; 
       (parse_con_suit env depth p) p' >] -> p'

and parse_con_suit env depth p = function
    [< 'Kwd "&"; 
       (parse_con env depth) p' >] -> Et (p,p')
|   [< >] -> p

and parse_dis env depth = function
    [< (parse_con env depth) p; 
       (parse_dis_suit env depth p) p' >] -> p'

and parse_dis_suit env depth p = function
    [< 'Kwd "|"; 
       (parse_dis env depth) p' >] -> Ou (p,p')
|   [< >] -> p

and parse_ar env depth = function
    [< (parse_dis env depth) p; 
       (parse_ar_suit env depth p) p' >] -> p'

and parse_ar_suit env depth p = function
    [< 'Kwd "->"; 
       (parse_ar env depth) p' >] -> Im (p,p')
|   [< 'Kwd "<->";
       (parse_ar env depth) p' >] -> Eq (p,p')
|   [< >] -> p

and parse_prop str = parse_ar [] 0 str
;;

let add_table s v =
  try myhashtbl__find name_tbl s;
    print_string ("\""^s^"\" all ready defined (ignored)\n")
  with Not_found ->
    myhashtbl__add name_tbl s v
;;


let parse_cmd = function
    [< 'Ident "bres"; parse_prop p; 'Kwd ";" >] -> 
      if res Binary p then print_string "proved\n"
                else print_string "not provable\n"
|   [< 'Ident "hres"; parse_prop p; 'Kwd ";" >] -> 
      if res Hyper p then print_string "proved\n"
                else print_string "not provable\n"
|   [< 'Ident "nhres"; parse_prop p; 'Kwd ";" >] -> 
      if res NegHyper p then print_string "proved\n"
                else print_string "not provable\n"
|   [< 'Ident "cst"; 'Ident s; 'Int n; 'Kwd ";" >] ->
      add_table s (Cst {Cst_name = s; Cst_arity = n; Cst_type = Fun_sym})
|   [< 'Ident "pred"; 'Ident s; 'Int n; 'Kwd ";" >] ->
      add_table s (Cst {Cst_name = s; Cst_arity = n; Cst_type = Pred_sym})
|   [< 'Ident "include"; 'String s ; 'Kwd ";" >] ->
      read_file s
|   [< 'Ident "axiom"; parse_prop p; 'Kwd ";" >] ->
      addax p
|   [< 'Ident "verbose"; 'Kwd ";" >] ->
      verbose := not ! verbose;
	  print_endline (if !verbose then "verbose on" else "verbose off")
|   [< 'Ident "reset"; 'Kwd ";" >] ->
      reset_ref (); myhashtbl__clear name_tbl; axiom := []
|   [< 'Ident "quit"; 'Kwd ";" >] -> raise Quit
;;

