(* Type des formules avec disjonctions et conjonctions n-aires au lieu
   de binaire, toutes les ngations tant ramenes au niveau atomique *)

type multi_prop =
  Dis of multi_prop list (* disjonction *)
| Con of multi_prop list (* conjonction *)
| pVar of int            (* variable *)
| nVar of int            (* ngation d'une variable *)
| sVrai | sFaux
;;

(* prop_to_multi_prop p transforme une formule de type prop en une
formule de type multi_prop. La traduction du connecteur Eq depend de
sa positivit dans la formule afin qu'une rgle binaire soit applique
en premier (donnat un branchement moindre dans la preuve *)

let prop_to_multi_prop p = 
  let rec fn pos b p =
    let rec fd l = fun
      true (Bin (Ou,p1,p2)) -> fd (fd l true p1) true p2
    | false (Bin (Et,p1,p2)) -> fd (fd l false p1) false p2
    | true (Bin (Im,p1,p2)) -> fd (fd l false p1) true p2
    | true (Bin (Eq,p1,p2)) -> fd (fd l true (Bin (Et,p1,p2))) 
                                  true (Bin (Et,Non p1,Non p2))
    | false (Bin (Eq,p1,p2)) -> fd (fd l false (Bin (Im,p1,p2)))
                                  false (Bin (Im,p2,p1))
    | b (Non p1) -> fd l (not b) p1
    | b p1 -> (gn (pos = b) p1)::l 
    in fd [] b p

  and gn pos = function
    Var n -> if pos then pVar n
                    else nVar n
  | Bin ((Ou|Im),_,_) as p -> let l = fn pos true p in
                       if pos then Dis l else Con l
  | Bin ((Et|Eq),_,_) as p -> let l = fn (not pos) false p in
                       if pos then Con l else Dis l
  | Non p -> gn (not pos) p
  | Vrai -> sVrai
  | Faux -> sFaux
  in gn true p
;;

(* test si une formule est une tautologie (aucune optimisation,  par
   l'application en premier des rgles unaires). Les variables
   rencontres dans le sequent sont placs dans une "map-table" tbl
   afin d'accelrer la recherche des axiomes *)

let taut_seq p =
  (* rgles unaires et axiomes *)
  let rec fd tbl l = function
    [] -> fc tbl l
  | (Dis l'')::l' -> fd tbl l (l''@l')
  | (Con _ as p)::l' -> fd tbl (p::l) l'
  | (pVar n)::l' -> (try 
                      let b = map__find n tbl in
                      if b then fd tbl l l' else true
                    with Not_found -> 
                      fd (map__add n true tbl) l l')
  | (nVar n)::l' -> (try 
                      let b = map__find n tbl in
                      if b then true else fd tbl l l'
                    with Not_found -> 
                      fd (map__add n false tbl) l l')
  | sVrai::l' -> true
  | sFaux::l' -> fd tbl l l'
  (* rgles binaires *)
  and fc tbl = function
    [] -> false
  | (Con l)::l' -> it_list (fun b p ->
                     if b then fd tbl l' [p]
                     else false) true l
  | _ -> raise (Failure "bug in sdecide")
  in fd (map__empty (prefix -)) [] [(prop_to_multi_prop p)]
;;
