
(* "vars" f fabrique la liste des variables de la formule f *)

let vars f = fn f
  where rec fn = function
    Var n -> [n]
  | Vrai | Faux -> []
  | Non f -> fn f
  | Bin(_,f,g) -> union (fn f) (fn g)
;;

(* "eval getv val f" teste si f est vraie dans la valuation val de type 'a en
   utilisant la fonction getv de type var -> 'a -> bool pour aller chercher
   la valeur d'une variable *)

let eval getv val f = fn f 
  where rec fn = function
    Var n -> getv n val
  | Vrai -> true
  | Faux -> false
  | Non f -> not (fn f)
  | Bin (c,f,g) -> match c with
      Et -> (fn f) & (fn g)
    | Ou -> (fn f) or (fn g)
    | Im -> not (fn f) or (fn g)
    | Eq -> (fn f) = (fn g)
;;

(* "taut init getv addone f" teste si la formule f est une tautologie en 
   utilisant la fonction "getv : var -> 'a -> bool" pour aller chercher la 
   valeur d'une variable, la fonction "addone : var -> bool -> 'a -> 'a" 
   pour ajouter un lment  une valuation et la 
   fonction "init : var list -> 'a" pour fabriquer la valuation initiale vide
*)

let taut init getv addone f = 
  let lv = vars f in
  fn (init lv) lv
  where rec fn val = function
    [] -> eval getv val f
  | n::l -> (fn (addone n true val) l) & (fn (addone n false val) l)
;;

(* test des tautologies avec des listes *)
let taut_list = taut (fun _ -> []) assoc (fun x b l -> (x,b)::l)
;;

(* test des tautologies avec des maps *)
let compare_int n p = if n < p then -1 else if n = p then 0 else 1;;
let taut_map = taut 
  (fun _ -> map__empty compare_int)
  map__find
  map__add
;;

(* test des tautologies avec des vecteurs *)
let taut_vect = taut 
  (fun lv -> make_vect ((it_list max 0 lv) + 1) false)
  (fun n v -> v.(n))
  (fun n b v -> v.(n) <- b; v)
;;
(* Attention : cette dernire version n'est pas purement fonctionnelle, et 
   sa correction dpend de l'ordre d'valuation *)
